// $Id$

/***************************************************************************
*   Copyright (C) 2004 by Thomas Fischer                                  *
*   fischer@unix-ag.uni-kl.de                                             *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <qfile.h>
#include <qtextstream.h>
#include <qstring.h>
#include <qvaluelist.h>

#include <bibtexfile.h>
#include <bibtexelement.h>
#include <bibtexcomment.h>
#include <bibtexentry.h>
#include <bibtexstring.h>
#include <bibtexpreamble.h>

const QString ExtraAlphaNumChars = QString( "?'`-_:.+/$\\" );

BibTeXFile::BibTeXFile( QObject *parent, const char *name ) : QObject( parent, name ), tokenChar( '\0' )
{}

BibTeXFile::~BibTeXFile()
{}

bool BibTeXFile::load( const QString& url )
{
    bool result = false;

    QFile input( url );
    if ( input.exists() && input.open( IO_ReadOnly ) )
    {
        QTextStream stream( &input );
        result = load( stream );
        input.close();
    }

    return result;
}

bool BibTeXFile::load( QTextStream& stream )
{
    elements.clear();
    cancelOp = false;

    Token token = readToken( stream );
    Token old_token = tUnknown;
    while ( token != tEOF && token != tUnknown && !cancelOp )
    {
        //qDebug( "token = %2i   tokenChar = #%i  tokenBuffer = '%s'", ( int ) token, tokenChar.latin1(), tokenBuffer.latin1() );

        switch ( token )
        {
        case tComment:
            // Warum überprüfen auf isEmpty?
            if ( tokenBuffer.isEmpty() )
                tokenBuffer = readChars( stream );
            elements.append( new BibTeXComment( tokenBuffer ) );
            break;

        case tPreamble:
            tokenBuffer = readChars( stream );
            elements.append( new BibTeXPreamble( tokenBuffer ) );
            break;

        case tString:
            {
                BibTeXString *string = readString( stream );
                if ( string != NULL )
                    elements.append( string );
                else
                    emit parseError( 1 );
            }
            break;

        case tEntry:
            {
                QString type = tokenBuffer;
                BibTeXEntry *entry = readEntry( type, stream );
                if ( entry != NULL )
                    elements.append( entry );
                else
                    emit parseError( 1 );
            }
            break;

        default:
            emit parseError( 0 );
        }

        old_token = token;
        token = readToken( stream );
        emit progress( stream.device() ->at(), stream.device() ->size() );
    }

    if ( cancelOp )
        elements.clear();

    return !cancelOp;
}

bool BibTeXFile::save( const QString& url )
{
    cancelOp = false;

    QFile output( url );
    if ( output.open( IO_WriteOnly ) )
    {
        QTextStream stream( &output );
        for ( ElementList::iterator it = elements.begin(); it != elements.end(); it++ )
        {
            stream << ( ( BibTeXElement* ) * it ) ->toBibTeXString() << endl << endl;
        }

        output.close();
    }

    return !cancelOp;
}

unsigned int BibTeXFile::count ()
{
    return elements.count();
}

BibTeXElement* BibTeXFile::at( const unsigned int index )
{
    return * ( elements.at( index ) );
}

BibTeXFile::Token BibTeXFile::readToken( QTextStream &stream )
{
    if ( ( tokenChar == '\0' ) || ( tokenChar.isSpace() ) )
    {
        stream.skipWhiteSpace();
        stream >> tokenChar;
    }

    if ( tokenChar == '@' )
    {
        stream >> tokenChar;
        tokenBuffer = "";
        while ( tokenChar.isLetter() )
        {
            tokenBuffer += tokenChar.lower();
            stream >> tokenChar;
        }

        if ( tokenBuffer == "comment" )
        {
            tokenBuffer = "";
            return tComment;
        }
        else if ( tokenBuffer == "preamble" )
        {
            tokenBuffer = "";
            return tPreamble;
        }
        else if ( tokenBuffer == "string" )
        {
            tokenBuffer = "";
            return tString;
        }
        else
            return tEntry;
    }
    else if ( ( tokenChar == '{' ) || ( tokenChar == '(' ) )
    {
        tokenChar = '\0';
        return tBraketOpen;
    }
    else if ( ( tokenChar == '}' ) || ( tokenChar == ')' ) )
    {
        tokenChar = '\0';
        return tBraketClose;
    }
    else if ( tokenChar == ',' )
    {
        tokenChar = '\0';
        return tComma;
    }
    else if ( tokenChar == '=' )
    {
        tokenChar = '\0';
        return tAssign;
    }
    else if ( tokenChar == '%' )
    {
        stream >> tokenChar;
        while ( ( tokenChar == '%' ) || tokenChar.isSpace() )
            stream >> tokenChar;

        tokenBuffer = stream.readLine();
        if ( ( tokenChar != '%' ) && !tokenChar.isSpace() )
            tokenBuffer = tokenChar + tokenBuffer;

        tokenChar = '\0';
        return tComment;
    }
    else if ( tokenChar == '#' )
    {
        tokenChar = '\0';
        return tStrConcat;
    }
    else if ( tokenChar.isLetterOrNumber() || ( ExtraAlphaNumChars.find( tokenChar ) >= 0 ) )
    {
        tokenBuffer = tokenChar;
        stream >> tokenChar;
        while ( tokenChar.isLetterOrNumber() || ( ExtraAlphaNumChars.find( tokenChar ) >= 0 ) )
        {
            tokenBuffer += tokenChar;
            stream >> tokenChar;
        }

        return tAlphaNumText;
    }

    if ( stream.atEnd() )
        return tEOF;

    qDebug( "Unexpected token char '%c' (%i) at stream position %li", tokenChar.latin1(), tokenChar.latin1(), stream.device() ->at() );
    stream >> tokenChar;

    emit parseError( 3 );

    return tUnknown;
}

void BibTeXFile::appendElement( BibTeXElement *element )
{
    elements.append( element );
}

void BibTeXFile::deleteElement( BibTeXElement *element )
{
    for ( ElementList::iterator it = elements.begin(); it != elements.end(); it++ )
        if ( *it == element )
        {
            elements.remove( it );
            break;
        }

    delete element;
}

QString BibTeXFile::readChars( QTextStream &stream )
{
    QString buffer = "";
    QChar c, old_c = '\0';
    int bracketCounter = 0;

    stream.skipWhiteSpace();
    stream >> c;

    if ( ( c == '{' ) || ( c == '(' ) )
    {
        char opening = c;
        char closing = ( c == '{' ) ? '}' : ')';

        bracketCounter = 1;
        old_c = c;
        stream >> c;
        while ( bracketCounter > 0 )
        {
            if ( c == closing )
                bracketCounter--;
            else if ( c == opening )
                bracketCounter++;
            if ( bracketCounter > 0 )
            {
                if ( c.latin1() < 16 )
                    c = ' ';
                if ( !c.isSpace() || !old_c.isSpace() )
                    buffer += c;
                old_c = c;
                stream >> c;
            }
        }
    }
    else if ( c == '"' )
    {
        old_c = c;
        stream >> c;
        while ( ( c != '"' ) || ( ( c == '"' ) && ( old_c == '\\' ) ) )
        {
            if ( c.latin1() < 16 )
                c = ' ';
            if ( !c.isSpace() || !old_c.isSpace() )
                buffer += c;
            old_c = c;
            stream >> c;
        }
    }
    else if ( c.isLetterOrNumber() || ( ExtraAlphaNumChars.find( c ) >= 0 ) )
    {
        buffer = c;
        old_c = c;
        stream >> c;
        while ( c.isLetterOrNumber() || ( ExtraAlphaNumChars.find( c ) >= 0 ) )
        {
            if ( c.latin1() < 16 )
                c = ' ';
            if ( !c.isSpace() || !old_c.isSpace() )
                buffer += c;
            old_c = c;
            stream >> c;
        }
        tokenChar = c;
    }

    //qDebug("readChars = %s", buffer.latin1());

    return buffer;
}

BibTeXString* BibTeXFile::readString( QTextStream &stream )
{
    BibTeXString * result = NULL;

    if ( ( readToken( stream ) == tBraketOpen ) && ( readToken( stream ) == tAlphaNumText ) )
    {
        QString id = tokenBuffer;
        if ( readToken( stream ) == tAssign )
        {
            QString value = readChars( stream );
            result = new BibTeXString( id, value );
        }
        while ( readToken( stream ) != tBraketClose )
            ;
    }

    return result;
}

BibTeXEntry* BibTeXFile::readEntry( QString &entryType, QTextStream &stream )
{
    BibTeXEntry * result = NULL;

    if ( ( readToken( stream ) == tBraketOpen ) && ( readToken( stream ) == tAlphaNumText ) )
    {
        result = new BibTeXEntry( BibTeXElement::elementTypeFromString( entryType ), tokenBuffer );

        QString fieldName;
        Token token = readToken( stream );
        while ( token != tBraketClose )
        {
            if ( token == tAlphaNumText )
            {
                fieldName = tokenBuffer;
            }
            else if ( token == tAssign )
            {
                QString value = readChars( stream );
                //qDebug("Entry: %s  %s", fieldName.latin1(), value.latin1());
                result->setField( fieldName, value );
            }
            token = readToken( stream );
        }

    }

    return result;
}

void BibTeXFile::cancelOperation()
{
    cancelOp = true;
}

#include "bibtexfile.moc"

